@@ -1,4 +1,5 @@
 #include <ruby.h>
+#include <stdarg.h>
 #ifdef HAVE_RUBY_ST_H
 #  include <ruby/st.h>
 #else
@@ -213,10 +214,16 @@ static const char *addr_any(sa_family_t family)
 	return ipv6;
 }
 
-static void bug_warn(void)
+static void bug_warn_nogvl(const char *fmt, ...)
 {
+	va_list ap;
+
+	va_start(ap, fmt);
+	vfprintf(stderr, fmt, ap);
+	va_end(ap);
+
 	fprintf(stderr, "Please report how you produced this at "\
-	                "raindrops@librelist.org\n");
+	                "raindrops-public@bogomips.org\n");
 	fflush(stderr);
 }
 
@@ -225,8 +232,8 @@ static struct listen_stats *stats_for(st_table *table, struct inet_diag_msg *r)
 	char *key, *port, *old_key;
 	size_t alloca_len;
 	struct listen_stats *stats;
-	size_t keylen;
-	size_t portlen = sizeof("65535");
+	socklen_t keylen;
+	socklen_t portlen = (socklen_t)sizeof("65535");
 	union any_addr sa;
 	socklen_t len = sizeof(struct sockaddr_storage);
 	int rc;
@@ -263,13 +270,12 @@ static struct listen_stats *stats_for(st_table *table, struct inet_diag_msg *r)
 		assert(0 && "unsupported address family, could that be IPv7?!");
 	}
 	if (rc != 0) {
-		fprintf(stderr, "BUG: getnameinfo: %s\n", gai_strerror(rc));
-		bug_warn();
+		bug_warn_nogvl("BUG: getnameinfo: %s\n", gai_strerror(rc));
 		*key = 0;
 	}
 
-	keylen = strlen(key);
-	portlen = strlen(port);
+	keylen = (socklen_t)strlen(key);
+	portlen = (socklen_t)strlen(port);
 
 	switch (sa.ss.ss_family) {
 	case AF_INET:
@@ -296,8 +302,7 @@ static struct listen_stats *stats_for(st_table *table, struct inet_diag_msg *r)
 				 addr_any(sa.ss.ss_family),
 				 ntohs(r->id.idiag_sport));
 		if (n <= 0) {
-			fprintf(stderr, "BUG: snprintf: %d\n", n);
-			bug_warn();
+			bug_warn_nogvl("BUG: snprintf: %d\n", n);
 		}
 		if (st_lookup(table, (st_data_t)key, (st_data_t *)&stats))
 			return stats;
@@ -391,8 +396,8 @@ static void prep_diag_args(
 
 	nladdr->nl_family = AF_NETLINK;
 
-	req->nlh.nlmsg_len = sizeof(struct diag_req) +
-	                    RTA_LENGTH(args->iov[2].iov_len);
+	req->nlh.nlmsg_len = (unsigned int)(sizeof(struct diag_req) +
+	                    RTA_LENGTH(args->iov[2].iov_len));
 	req->nlh.nlmsg_type = TCPDIAG_GETSOCK;
 	req->nlh.nlmsg_flags = NLM_F_ROOT | NLM_F_MATCH | NLM_F_REQUEST;
 	req->nlh.nlmsg_pid = getpid();
@@ -465,12 +470,12 @@ static VALUE diag(void *ptr)
 		}
 	}
 out:
-	{
+	/* prepare to raise, free memory before reacquiring GVL */
+	if (err && args->table) {
 		int save_errno = errno;
-		if (err && args->table) {
-			st_foreach(args->table, st_free_data, 0);
-			st_free_table(args->table);
-		}
+
+		st_foreach(args->table, st_free_data, 0);
+		st_free_table(args->table);
 		errno = save_errno;
 	}
 	return (VALUE)err;
@@ -584,6 +589,10 @@ static void gen_bytecode(struct iovec *iov, union any_addr *inet)
 	}
 }
 
+/*
+ * n.b. we may safely raise here because an error will cause diag()
+ * to free args->table
+ */
 static void nl_errcheck(VALUE r)
 {
 	const char *err = (const char *)r;
@@ -609,6 +618,13 @@ static VALUE tcp_stats(struct nogvl_args *args, VALUE addr)
 	return rb_listen_stats(&args->stats);
 }
 
+static int drop_placeholders(st_data_t k, st_data_t v, st_data_t ign)
+{
+	if ((VALUE)v == Qtrue)
+		return ST_DELETE;
+	return ST_CONTINUE;
+}
+
 /*
  * call-seq:
  *      Raindrops::Linux.tcp_listener_stats([addrs[, sock]]) => hash
@@ -649,10 +665,9 @@ static VALUE tcp_listener_stats(int argc, VALUE *argv, VALUE self)
 	case T_ARRAY: {
 		long i;
 		long len = RARRAY_LEN(addrs);
-		VALUE cur;
 
 		if (len == 1) {
-			cur = rb_ary_entry(addrs, 0);
+			VALUE cur = rb_ary_entry(addrs, 0);
 
 			rb_hash_aset(rv, cur, tcp_stats(&args, cur));
 			return rv;
@@ -662,7 +677,7 @@ static VALUE tcp_listener_stats(int argc, VALUE *argv, VALUE self)
 			VALUE cur = rb_ary_entry(addrs, i);
 
 			parse_addr(&check, cur);
-			rb_hash_aset(rv, cur, Qtrue);
+			rb_hash_aset(rv, cur, Qtrue /* placeholder */);
 		}
 		/* fall through */
 	}
@@ -680,6 +695,9 @@ static VALUE tcp_listener_stats(int argc, VALUE *argv, VALUE self)
 	st_foreach(args.table, NIL_P(addrs) ? st_to_hash : st_AND_hash, rv);
 	st_free_table(args.table);
 
+	if (RHASH_SIZE(rv) > 1)
+		rb_hash_foreach(rv, drop_placeholders, Qfalse);
+
 	/* let GC deal with corner cases */
 	if (argc < 2) rb_io_close(sock);
 	return rv;
